from typing import Optional
import shutil
import os


def get_pager() -> Optional[str]:
    # Windows comes pre-installed with the 'more' pager.
    # See https://superuser.com/a/426229
    # Unix distributions also have 'more' pre-installed.
    more_bin = shutil.which("more")
    less_bin = shutil.which("less")
    default_pager = more_bin
    if less_bin is not None:
        default_pager = less_bin
    pager = os.getenv("PAGER", default_pager)
    return pager



import unittest
from unittest.mock import patch
from argparse import ArgumentError, Namespace
from datetime import datetime, timezone, timedelta, date

import worklog.constants as wc
from worklog.dispatcher import dispatch


@patch("configparser.ConfigParser")
@patch("argparse.ArgumentParser")
@patch("worklog.log")
class TestDispatchCommit(unittest.TestCase):
    def test_invalid(self, mock_log, mock_parser, mock_cfg):
        ns = Namespace(
            subcmd="commit",
            type="invalid_value",
            offset_minutes=0,
            time=None,
            force=False,
        )
        dispatch(mock_log, mock_parser, ns, mock_cfg)

        mock_log.commit.assert_not_called()

    def test_start(self, mock_log, mock_parser, mock_cfg):
        ns = Namespace(
            subcmd="commit", type="start", offset_minutes=0, time=None, force=False
        )
        dispatch(mock_log, mock_parser, ns, mock_cfg)

        mock_log.commit.assert_called_once_with(
            "session", "start", 0, None, force=False
        )

    def test_stop(self, mock_log, mock_parser, mock_cfg):
        ns = Namespace(
            subcmd="commit", type="stop", offset_minutes=0, time=None, force=False
        )
        dispatch(mock_log, mock_parser, ns, mock_cfg)

        mock_log.commit.assert_called_once_with("session", "stop", 0, None, force=False)

    def test_start_with_offset(self, mock_log, mock_parser, mock_cfg):
        ns = Namespace(
            subcmd="commit", type="stop", offset_minutes=-11, time=None, force=False
        )
        dispatch(mock_log, mock_parser, ns, mock_cfg)

        mock_log.commit.assert_called_once_with(
            "session", "stop", -11, None, force=False
        )

    def test_start_with_time(self, mock_log, mock_parser, mock_cfg):
        ns = Namespace(
            subcmd="commit",
            type="stop",
            offset_minutes=0,
            time=datetime(2020, 1, 1, tzinfo=timezone.utc),
            force=False,
        )
        dispatch(mock_log, mock_parser, ns, mock_cfg)

        mock_log.commit.assert_called_once_with(
            "session", "stop", 0, datetime(2020, 1, 1, tzinfo=timezone.utc), force=False
        )

    def test_start_forced(self, mock_log, mock_parser, mock_cfg):
        ns = Namespace(
            subcmd="commit", type="stop", offset_minutes=0, time=None, force=True,
        )
        dispatch(mock_log, mock_parser, ns, mock_cfg)

        mock_log.commit.assert_called_once_with("session", "stop", 0, None, force=True)


@patch("configparser.ConfigParser")
@patch("argparse.ArgumentParser")
@patch("worklog.log")
class TestDispatchTask(unittest.TestCase):
    def test_start(self, mock_log, mock_parser, mock_cfg):
        """It should be possible to start a new task"""
        ns = Namespace(
            subcmd="task",
            type="start",
            auto_close=False,
            offset_minutes=0,
            time=None,
            id="foobar",
        )
        dispatch(mock_log, mock_parser, ns, mock_cfg)

        mock_log.stop_active_tasks.assert_not_called()
        mock_log.commit.assert_called_once_with(
            "task", "start", 0, None, identifier="foobar"
        )

    def test_stop(self, mock_log, mock_parser, mock_cfg):
        """It should be possible to stop a task"""
        ns = Namespace(
            subcmd="task",
            type="stop",
            auto_close=False,
            offset_minutes=0,
            time=None,
            id="foobar",
        )
        dispatch(mock_log, mock_parser, ns, mock_cfg)

        mock_log.stop_active_tasks.assert_not_called()
        mock_log.commit.assert_called_once_with(
            "task", "stop", 0, None, identifier="foobar"
        )

    def test_start_with_auto_close(self, mock_log, mock_parser, mock_cfg):
        """
        It should be possible to start a new task and auto close all
        currently runnning tasks.
        """
        ns = Namespace(
            subcmd="task",
            type="start",
            auto_close=True,
            offset_minutes=0,
            time=None,
            id="foobar",
        )
        dispatch(mock_log, mock_parser, ns, mock_cfg)

        mock_log.stop_active_tasks.assert_called_once()
        mock_log.commit.assert_called_once_with(
            "task", "start", 0, None, identifier="foobar"
        )

    def test_list_tasks(self, mock_log, mock_parser, mock_cfg):
        """It should be possible to list tasks."""
        ns = Namespace(subcmd="task", type="list")
        dispatch(mock_log, mock_parser, ns, mock_cfg)

        mock_log.list_tasks.assert_called_once()

    def test_report_task_by_id(self, mock_log, mock_parser, mock_cfg):
        """It should be possible to report infos about a single task id."""
        ns = Namespace(subcmd="task", type="report", id="foobar")
        dispatch(mock_log, mock_parser, ns, mock_cfg)

        mock_log.task_report.assert_called_once_with("foobar")


@patch("configparser.ConfigParser")
@patch("argparse.ArgumentParser")
@patch("worklog.log")
class TestDispatchDoctor(unittest.TestCase):
    def test_doctor(self, mock_log, mock_parser, mock_cfg):
        ns = Namespace(subcmd="doctor")
        dispatch(mock_log, mock_parser, ns, mock_cfg)

        mock_log.doctor.assert_called_once()


@patch("configparser.ConfigParser")
@patch("argparse.ArgumentParser")
@patch("worklog.log")
class TestDispatchReport(unittest.TestCase):
    def test_report(self, mock_log, mock_parser, mock_cfg):
        date_from = datetime(2020, 1, 1, tzinfo=timezone.utc)
        date_to = datetime(2020, 1, 2, tzinfo=timezone.utc)
        ns = Namespace(subcmd="report", date_from=date_from, date_to=date_to)
        dispatch(mock_log, mock_parser, ns, mock_cfg)

        mock_log.report.assert_called_once_with(date_from, date_to)


@patch("configparser.ConfigParser")
@patch("argparse.ArgumentParser")
@patch("worklog.log")
class TestDispatchStatus(unittest.TestCase):
    def test_status(self, mock_log, mock_parser, mock_cfg):
        mock_cfg.get.side_effect = ["8.0", "10.0"]
        ns = Namespace(subcmd="status", fmt=None, yesterday=False, date=None)
        dispatch(mock_log, mock_parser, ns, mock_cfg)

        expected_query_date = datetime.today().date()

        mock_log.status.assert_called_once_with(
            8.0, 10.0, fmt=None, query_date=expected_query_date
        )

    def test_status_yesterday(self, mock_log, mock_parser, mock_cfg):
        mock_cfg.get.side_effect = ["8.0", "10.0"]
        ns = Namespace(subcmd="status", fmt=None, yesterday=True, date=None)
        dispatch(mock_log, mock_parser, ns, mock_cfg)

        expected_query_date = datetime.today().date() - timedelta(days=1)

        mock_log.status.assert_called_once_with(
            8.0, 10.0, fmt=None, query_date=expected_query_date
        )

    def test_status_fixed_date(self, mock_log, mock_parser, mock_cfg):
        mock_cfg.get.side_effect = ["8.0", "10.0"]
        ns = Namespace(
            subcmd="status",
            fmt=None,
            yesterday=False,
            date=datetime(2020, 1, 1, tzinfo=timezone.utc),
        )
        dispatch(mock_log, mock_parser, ns, mock_cfg)

        expected_query_date = date(2020, 1, 1)

        mock_log.status.assert_called_once_with(
            8.0, 10.0, fmt=None, query_date=expected_query_date
        )


@patch("configparser.ConfigParser")
@patch("argparse.ArgumentParser")
@patch("worklog.log")
class TestDispatchLog(unittest.TestCase):
    def test_log_default(self, mock_log, mock_parser, mock_cfg):
        mock_cfg.get.side_effect = ["10"]  # worklog.no_pager_max_entries
        ns = Namespace(subcmd="log", no_pager=False, all=False, category=None, number=5)
        dispatch(mock_log, mock_parser, ns, mock_cfg)

        mock_log.log.assert_called_once_with(5, False, None)

    def test_log_all(self, mock_log, mock_parser, mock_cfg):
        mock_cfg.get.side_effect = ["10"]  # worklog.no_pager_max_entries
        ns = Namespace(subcmd="log", no_pager=False, all=True, category=None, number=5)
        dispatch(mock_log, mock_parser, ns, mock_cfg)

        mock_log.log.assert_called_once_with(-1, True, None)

    def test_log_large_n(self, mock_log, mock_parser, mock_cfg):
        """It should automatically use pager if number > no_pager_max_entries"""
        mock_cfg.get.side_effect = ["10"]  # worklog.no_pager_max_entries
        ns = Namespace(
            subcmd="log", no_pager=False, all=False, category=None, number=20
        )
        dispatch(mock_log, mock_parser, ns, mock_cfg)

        mock_log.log.assert_called_once_with(20, True, None)



import unittest
import numpy as np

from worklog.utils.schema import empty_df_from_schema, get_datetime_cols_from_schema


class TestSchema(unittest.TestCase):
    def test_empty_df(self):
        schema = [
            ("commit_dt", "datetime64[ns]",),
            ("log_dt", "datetime64[ns]",),
            ("category", "object",),
            ("type", "object",),
        ]

        df = empty_df_from_schema(schema)
        self.assertListEqual(
            df.dtypes.values.tolist(),
            [np.dtype("<M8[ns]"), np.dtype("<M8[ns]"), np.dtype("O"), np.dtype("O")],
        )
        self.assertTupleEqual(df.shape, (0, len(schema)))

    def test_get_datetime_cols_from_schema(self):
        schema = [
            ("commit_dt", "datetime64[ns]",),
            ("log_dt", "datetime64[ns]",),
            ("category", "object",),
            ("type", "object",),
        ]

        actual = get_datetime_cols_from_schema(schema)
        self.assertListEqual(["commit_dt", "log_dt"], actual)



from typing import Union, Optional
from datetime import timedelta
import pandas as pd
import numpy as np
from math import floor


def format_timedelta(value: Optional[Union[timedelta, np.timedelta64]]) -> str:
    if value is None or value is pd.NaT or value is np.nan:
        return "{:02}:{:02}:{:02}".format(0, 0, 0)
    elif isinstance(value, timedelta):
        return _format_timedelta_py(value)
    elif isinstance(value, np.timedelta64):
        return _format_timedelta_np(value)
    else:
        raise ValueError("value must be either a Python or a numpy timedelta instance")


def _format_timedelta_py(td: timedelta) -> str:
    try:
        total_secs = td.total_seconds()
        hours, remainder = divmod(total_secs, 3600)
        minutes, seconds = divmod(remainder, 60)
        return "{:02}:{:02}:{:02}".format(int(hours), int(minutes), int(seconds))
    except (AttributeError, ValueError):
        return "{:02}:{:02}:{:02}".format(0, 0, 0)


def _format_timedelta_np(value: np.timedelta64) -> str:
    seconds = value / np.timedelta64(1, "s")
    if np.isnan(seconds):
        hours = minutes = seconds = 0
    else:
        hours = floor(seconds / 3600)
        minutes = floor((seconds - hours * 3600) / 60)
        seconds = floor(seconds % 60)
    return "{hours:02}:{minutes:02}:{seconds:02}".format(
        hours=hours, minutes=minutes, seconds=seconds
    )



from typing import List
from datetime import datetime, timezone, timedelta
import re
import argparse

import worklog.constants as wc

_help_time_arg = (
    "Exact point in time. "
    "Can be a either hours and minutes (format: 'hh:mm') on the same day or a full ISO "
    "format string, such as '2020-08-05T08:15:00+02:00'. "
    "In the latter case the local timezone is used if the timezone part is left empty."
)


def get_arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        "Worklog", description="Simple CLI tool to log work and projects."
    )
    parser.add_argument("-v", "--verbose", dest="verbosity", action="count", default=0)

    subparsers = parser.add_subparsers(dest="subcmd")

    _add_commit_parser(subparsers)
    _add_task_parser(subparsers)
    _add_status_parser(subparsers)
    _add_doctor_parser(subparsers)
    _add_log_parser(subparsers)
    _add_report_parser(subparsers)

    return parser


def _add_commit_parser(subparsers: argparse._SubParsersAction):
    commit_parser = subparsers.add_parser(
        wc.SUBCMD_COMMIT,
        description=(
            "Commit the start or end of a new working session to the worklog file. "
            "Use this function to stamp in the morning and stamp out in the evening."
        ),
    )
    commit_parser.add_argument(
        "type",
        choices=[wc.TOKEN_START, wc.TOKEN_STOP],
        help="Persists a new work session or closes a work session.",
    )
    _add_timeshift_args(commit_parser)
    commit_parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Force command, will auto-stop running tasks.",
    )


def _add_task_parser(subparsers: argparse._SubParsersAction):
    task_parser = subparsers.add_parser(
        wc.SUBCMD_TASK,
        description=(
            "Tasks are pieces of work to be done or undertaken. "
            "A task can only be started during an ongoing session. "
            "Use 'wl commit start' to start a new working session."
        ),
    )
    task_parser_type = task_parser.add_subparsers(dest="type")

    # task start
    task_start_parser = task_parser_type.add_parser(wc.TOKEN_START)
    task_start_parser.add_argument(
        "id", type=str, help="Task identifier, can be freely chosen",
    )
    task_start_parser.add_argument(
        "-ac", "--auto-close", action="store_true", help=("Auto closes open tasks."),
    )
    _add_timeshift_args(task_start_parser)

    # task stop
    task_stop_parser = task_parser_type.add_parser(wc.TOKEN_STOP)
    task_stop_parser.add_argument(
        "id", type=str, help="Task identifier of a running task.",
    )
    _add_timeshift_args(task_stop_parser)

    # task list
    task_list_parser = task_parser_type.add_parser("list")

    # task report
    task_report_parser = task_parser_type.add_parser("report")
    task_report_parser.add_argument(
        "id", type=str, help="Task identifier of a recorded task.",
    )


def _add_status_parser(subparsers: argparse._SubParsersAction):
    status_parser = subparsers.add_parser(
        wc.SUBCMD_STATUS,
        description=(
            "The status commend shows the tracking results for an individual day. "
            "By default the current day is selected. "
            "For integration of the current status into a status bar, use the `--fmt` "
            "argument."
        ),
    )
    status_time_grp = status_parser.add_mutually_exclusive_group()
    status_time_grp.add_argument(
        "--yesterday",
        action="store_true",
        help="Returns the status of yesterday instead of today.",
    )
    status_time_grp.add_argument(
        "--date", type=_year_month_day_parser, help=("Date in the form YYYY-MM-DD")
    )
    status_parser.add_argument(
        "--fmt", type=str, default=None, help="Use a custom formatted string"
    )


def _add_doctor_parser(subparsers: argparse._SubParsersAction):
    doctor_parser = subparsers.add_parser(
        wc.SUBCMD_DOCTOR,
        description=(
            "The doctor command checks the worklog for missing or problematic entries. "
            "It will report the following issues: non-closed working sessions"
        ),
    )


def _add_log_parser(subparsers: argparse._SubParsersAction):
    log_parser = subparsers.add_parser(
        wc.SUBCMD_LOG,
        description=(
            "Shows the content of the worklog file sorted after the date and time of the "
            "entry. "
            "Use this command to manually review the content of the worklog."
        ),
    )
    log_parser.add_argument(
        "-n",
        "--number",
        type=_positive_int,
        default=10,
        help=(
            "Defines many log entries should be shown. "
            "System pager will be used if n > 20."
        ),
    )
    log_parser.add_argument(
        "-a",
        "--all",
        action="store_true",
        help="Show all entries. System pager will be used.",
    )
    log_parser.add_argument(
        "--category", choices=[wc.TOKEN_SESSION, wc.TOKEN_TASK], help="Filter category",
    )
    log_parser.add_argument(
        "--no-pager",
        action="store_true",
        help=(
            "Don't use a the system pager. "
            "Prints all output to STDOUT regardless of how many entries will be shown. "
            "This flag should be used if there are problems with the system pager."
        ),
    )


def _add_report_parser(subparsers: argparse._SubParsersAction):
    now = datetime.now(timezone.utc).astimezone().replace(microsecond=0)
    current_month: str = now.replace(day=1).isoformat()[: len("2000-01-01")]
    next_month: str = (now.replace(day=1) + timedelta(days=31)).replace(
        day=1
    ).isoformat()[: len("2000-01-01")]

    report_parser = subparsers.add_parser(
        wc.SUBCMD_REPORT,
        description=(
            "Creates a report for a given time window. "
            "Working time will be aggregated on a monthly, weekly and daily basis. "
            "Tasks will be aggregated separately. "
            "By default the current month will be used for the report."
        ),
    )
    report_parser.add_argument(
        "--date-from",
        type=_combined_month_or_day_or_week_parser,
        default=current_month,
        help=(
            "Date from which the aggregation should be started (inclusive). "
            "By default the start of the current calendar month is selected. "
            "Allowed input formats are YYYY-MM-DD, YYYY-MM and YYYY-WXX, with "
            "XX referring to the week number, e.g. 35."
        ),
    )
    report_parser.add_argument(
        "--date-to",
        type=_combined_month_or_day_or_week_parser,
        default=next_month,
        help=(
            "Date to which the aggregation should be started (exclusive). "
            "By default the next calendar month is selected. "
            "Allowed input formats are YYYY-MM-DD, YYYY-MM and YYYY-WXX, with "
            "XX referring to the week number, e.g. 35."
        ),
    )


def _combined_month_or_day_or_week_parser(value: str) -> datetime:
    if re.match(r"^\d{4}\-\d{2}$", value):
        return _year_month_parser(value)
    elif re.match(r"^\d{4}\-\d{2}\-\d{2}$", value):
        return _year_month_day_parser(value)
    elif re.match(r"^\d{4}-W\d{2}$", value):
        return _calendar_week_parser(value)
    raise argparse.ArgumentTypeError(f"{value} is not a valid format")


def _year_month_parser(value: str) -> datetime:
    if not re.match(r"^\d{4}\-\d{2}$", value):
        raise argparse.ArgumentTypeError(f"{value} is not in the format YYYY-MM")
    year, month = [int(x) for x in value.split("-")]
    return datetime(year=year, month=month, day=1, tzinfo=wc.LOCAL_TIMEZONE)


def _year_month_day_parser(value: str) -> datetime:
    if not re.match(r"^\d{4}\-\d{2}\-\d{2}$", value):
        raise argparse.ArgumentTypeError(f"{value} is not in the format YYYY-MM-DD")
    year, month, day = [int(x) for x in value.split("-")]
    return datetime(year=year, month=month, day=day, tzinfo=wc.LOCAL_TIMEZONE)


def _calendar_week_parser(value: str) -> datetime:
    if not re.match(r"^\d{4}-W\d{2}$", value):
        raise argparse.ArgumentTypeError(f"{value} is not in the format cwWW")
    dt = datetime.strptime(value + "-1", "%Y-W%W-%w").replace(tzinfo=wc.LOCAL_TIMEZONE)
    return dt


def _positive_int(value: str) -> int:
    value_int = int(value)
    if value_int <= 0:
        raise argparse.ArgumentTypeError(f"{value} is not a positive int value.")
    return value_int


def _add_timeshift_args(parser: argparse.ArgumentParser):
    timeshift_grp = parser.add_mutually_exclusive_group()
    timeshift_grp.add_argument(
        "-om",
        "--offset-minutes",
        type=float,
        default=0,
        help=(
            "Offset of the start/stop time in minutes. "
            "Positive values shift the timestamp into the future, negative "
            "values shift it into the past."
        ),
    )
    timeshift_grp.add_argument("-t", "--time", help=_help_time_arg)




from typing import Optional
from pandas import DataFrame
from datetime import datetime, date, timezone, tzinfo
import logging

import worklog.constants as wc


def check_order_session(df_group: DataFrame, logger: logging.Logger):
    last_type = None
    for i, row in df_group.where(
        df_group[wc.COL_CATEGORY] == wc.TOKEN_SESSION
    ).iterrows():
        if i == 0 and row[wc.COL_TYPE] != wc.TOKEN_START:
            logger.error(
                f'First entry of type "{wc.TOKEN_SESSION}" on date {row.date} is not "{wc.TOKEN_START}".'
            )
        if row[wc.COL_TYPE] == last_type:
            logger.error(
                f'"{wc.TOKEN_SESSION}" entries on date {row.date} are not ordered correctly.'
            )
        last_type = row[wc.COL_TYPE]
    if last_type != wc.TOKEN_STOP:
        logger.error(f"Date {row.date} has no stop entry.")


def sentinel_datetime(
    target_date: date, tzinfo: Optional[tzinfo] = wc.LOCAL_TIMEZONE
) -> datetime:
    if target_date > datetime.now().date():
        raise ValueError("Only dates on the same day or in the past are supported.")
    return min(
        datetime.now(timezone.utc).astimezone(tz=tzinfo).replace(microsecond=0),
        datetime(
            target_date.year, target_date.month, target_date.day, 23, 59, 59, 0, tzinfo,
        ).astimezone(tz=tzinfo),
    )


def is_active_session(df: DataFrame):
    """
    Returns True if the last entry in a given pandas DataFrame has the
    category 'session' and type 'start'.

    Note: Make sure to apply this method only on a sorted DataFrame.
    """
    return df.iloc[-1][wc.COL_TYPE] == wc.TOKEN_START if df.shape[0] > 0 else False



import logging
import subprocess
import sys
import tempfile
from datetime import date, datetime, timedelta, timezone
from io import StringIO
from math import floor
from pathlib import Path
from typing import List, Optional, Tuple

import numpy as np  # type: ignore
import pandas as pd  # type: ignore

from worklog.breaks import AutoBreak
import worklog.constants as wc
from worklog.utils.pager import get_pager
from worklog.utils.time import calc_log_time, extract_date_and_time
from worklog.utils.schema import empty_df_from_schema, get_datetime_cols_from_schema
from worklog.utils.formatting import format_timedelta
from worklog.utils.tasks import (
    calc_task_durations,
    extract_intervals,
    get_active_task_ids,
    get_all_task_ids_with_duration,
)
from worklog.utils.session import (
    check_order_session,
    sentinel_datetime,
    is_active_session,
)

logger = logging.getLogger(wc.DEFAULT_LOGGER_NAME)


class Log(object):
    # In-memory representation of log
    _log_df: pd.DataFrame = None

    # Backend file config
    _log_fp: Optional[str] = None
    _separator: Optional[str] = None
    _schema: List[Tuple[str, str]] = [
        (wc.COL_COMMIT_DATETIME, "datetime64[ns]",),
        (wc.COL_LOG_DATETIME, "datetime64[ns]",),
        (wc.COL_CATEGORY, "object",),
        (wc.COL_TYPE, "object",),
        (wc.COL_TASK_IDENTIFIER, "object",),
    ]

    # Error messages
    _err_msg_empty_log = (
        "Fatal: No log data available. Start a new log entry with 'wl commit start'.\n"
    )
    _err_msg_empty_log_short = "N/A"
    _err_msg_log_data_missing_for_date = "No log data available for {query_date}.\n"
    _err_msg_log_data_missing_for_date_short = "N/A"
    _err_msg_commit_active_tasks = (
        "Fatal. Cannot stop, because tasks are still running. "
        "Stop running tasks first: {active_tasks:} or use --force flag.\n"
    )

    auto_break: AutoBreak = AutoBreak()

    def __init__(self, fp: str, separator: str = "|") -> None:
        self._log_fp = fp
        self._separator = separator

        Path(self._log_fp).touch(mode=0o660)
        self._read()

    def commit(
        self,
        category: str,
        type_: str,
        offset_min: int = 0,
        time: Optional[str] = None,
        identifier: str = None,
        force: bool = False,
    ) -> None:
        """Commit a session/task change to the logfile."""
        log_date = calc_log_time(offset_min, time)
        self._commit(category, type_, log_date, identifier, force)

    def doctor(self) -> None:
        """Test if the logfile is consistent."""
        self._log_df.groupby(["date"]).apply(
            lambda group: check_order_session(group, logger)
        )

    def list_tasks(self):
        """List all known tasks, i.e. tasks that have been used previously
        and are stored in the logfile."""
        task_df = self._log_df[self._log_df[wc.COL_CATEGORY] == wc.TOKEN_TASK]
        sys.stdout.write("These tasks are listed in the log:\n")
        for task_id in sorted(task_df[wc.COL_TASK_IDENTIFIER].unique()):
            sys.stdout.write(f"{task_id}\n")

    def log(
        self, n: int, use_pager: bool, filter_category: Optional[List[str]]
    ) -> None:
        """Display the content of the logfile."""
        if self._log_df.shape[0] == 0:
            sys.stdout.write("No data available\n")
            return

        fields = ["date", "time", wc.COL_CATEGORY, wc.COL_TYPE, wc.COL_TASK_IDENTIFIER]
        df = self._log_df[fields].iloc[::-1]  # sort in reverse (latest first)
        df[wc.COL_TASK_IDENTIFIER] = df[wc.COL_TASK_IDENTIFIER].fillna("-")
        if filter_category:
            df = df[df[wc.COL_CATEGORY] == filter_category]
        if n > 0:
            df = df.head(n=n)
        if not use_pager:
            sys.stdout.write(df.to_string(index=False) + "\n")
        else:
            with tempfile.NamedTemporaryFile(mode="w") as fh:
                logger.debug(f"Write content to temporary file: {fh.name}")
                fh.write(df.to_string(index=False))
                fh.flush()
                pager = get_pager()
                if pager is None:
                    sys.stdout.write(df.to_string(index=False) + "\n")
                else:
                    logger.debug(f"Set pager to {pager}")
                    process = subprocess.Popen([pager, fh.name])
                    process.wait()

    def report(self, date_from: datetime, date_to: datetime):
        """Generate a daily, weekly, monthly and task based report based on
        the content in the logfile."""
        session_mask = self._log_df[wc.COL_CATEGORY] == wc.TOKEN_SESSION
        task_mask = self._log_df[wc.COL_CATEGORY] == wc.TOKEN_TASK
        time_mask = (self._log_df[wc.COL_LOG_DATETIME] >= date_from) & (
            self._log_df[wc.COL_LOG_DATETIME] < date_to
        )

        # Day aggregation
        df_day = self._aggregate_time(time_mask & session_mask, resample="D")
        df_day["break"] = df_day["agg_time"].map(self.auto_break.get_duration)

        # Week aggregation
        df_week = (
            df_day.set_index(wc.COL_LOG_DATETIME).resample("W").sum().reset_index()
        )

        # Month aggregration
        df_month = (
            df_day.set_index(wc.COL_LOG_DATETIME).resample("M").sum().reset_index()
        )

        for df in (df_day, df_week, df_month):
            df["agg_time_bookable"] = df["agg_time"] - df["break"]

        # Task aggregation
        df_tasks = self._aggregate_tasks(time_mask & task_mask)

        print_cols = [wc.COL_LOG_DATETIME, "agg_time"]
        print_cols_labels = ["Date", "Total time"]
        if self.auto_break.active:
            print_cols += ["break", "agg_time_bookable"]
            print_cols_labels += ["Break", "Bookable time"]

        def _formatters(date_type: str = "M"):
            date_max_len = len("2000-01") if date_type == "M" else len("2000-01-01")
            return {
                wc.COL_LOG_DATETIME: lambda v: str(v.date())[:date_max_len],
                "agg_time": format_timedelta,
                "agg_time_bookable": format_timedelta,
            }

        self._print_aggregation(
            "month",
            df_month,
            print_cols,
            print_cols_labels,
            formatters=_formatters("M"),
        )
        self._print_aggregation(
            "week", df_week, print_cols, print_cols_labels, formatters=_formatters("D"),
        )
        self._print_aggregation(
            "day", df_day, print_cols, print_cols_labels, formatters=_formatters("D")
        )

        print_cols = [wc.COL_TASK_IDENTIFIER, "agg_time"]
        print_cols_labels = ["Task name", "Total time"]
        self._print_aggregation(
            "tasks",
            df_tasks,
            print_cols,
            print_cols_labels,
            formatters=_formatters("D"),
        )

    def status(
        self, hours_target: float, hours_max: float, query_date: date, fmt: str = None,
    ) -> None:
        """Display the current working status, e.g. total time worked at this
        day, remaining time, etc."""
        self._check_nonempty_or_exit(fmt)

        df_day = self._filter_date_category_limit_cols(query_date)

        if df_day.shape[0] == 0:
            if fmt is None:
                msg = self._err_msg_log_data_missing_for_date.format(
                    query_date=query_date
                )
                sys.stderr.write(msg)
            else:
                sys.stdout.write(self._err_msg_log_data_missing_for_date_short)
            return

        is_active = is_active_session(df_day)
        logger.debug(f"Is active: {is_active}")

        df_day = self._add_sentinel(query_date, df_day)
        facts = self._calc_facts(df_day, hours_target, hours_max)

        date_mask = self._log_df["date"] == query_date
        task_mask = self._log_df[wc.COL_CATEGORY] == wc.TOKEN_TASK
        sel_task_mask = date_mask & task_mask
        all_touched_tasks = get_all_task_ids_with_duration(self._log_df[sel_task_mask])
        active_tasks = get_active_task_ids(self._log_df[sel_task_mask])

        lines = [
            ("Status", "Tracking {status}"),
            ("Total time", "{total_time} ({percentage:3}%)"),
            ("Remaining time", "{remaining_time} ({percentage_remaining:3}%)"),
            ("Overtime", "{overtime} ({percentage_overtime:3}%)"),
            ("Break Duration", "{break_duration}"),
            ("All touched tasks", "{all_touched_tasks}",),
            ("Active tasks", "{active_tasks}",),
        ]

        if is_active and date == "today":
            lines += [("End of work", facts["end_time_str"],)]

        key_max_len = max([len(line[0]) for line in lines])
        fmt_string = "{:" + str(key_max_len + 1) + "s}: {}"

        stdout_fmt = "\n".join(fmt_string.format(*line) for line in lines) + "\n"

        sys.stdout.write(
            (stdout_fmt if fmt is None else fmt).format(
                **facts,
                status="on" if is_active else "off",
                active_tasks=f"({len(active_tasks)}) [" + ", ".join(active_tasks) + "]",
                all_touched_tasks=f"({len(all_touched_tasks)}) ["
                + ", ".join(
                    [
                        f"{k} ({format_timedelta(v)})"
                        for k, v in all_touched_tasks.items()
                    ]
                )
                + "]",
            )
        )

    def stop_active_tasks(self, log_dt: datetime):
        """Stop all active tasks by commiting changes to the logfile."""
        query_date = log_dt.date()
        task_mask = self._log_df[wc.COL_CATEGORY] == wc.TOKEN_TASK
        date_mask = self._log_df["date"] == query_date
        mask = task_mask & date_mask
        active_task_ids = get_active_task_ids(self._log_df[mask])
        for task_id in active_task_ids:
            self._commit(wc.TOKEN_TASK, wc.TOKEN_STOP, log_dt, identifier=task_id)

    def task_report(self, task_id):
        """Generate a report of a given task."""
        task_mask = self._log_df[wc.COL_CATEGORY] == wc.TOKEN_TASK
        task_id_mask = self._log_df[wc.COL_TASK_IDENTIFIER] == task_id
        mask = task_mask & task_id_mask
        task_df = self._log_df[mask]

        if task_df.shape[0] == 0:
            sys.stderr.write(
                (
                    f"Task ID {task_id} is unknown. "
                    "See 'wl task list' to list all known tasks.\n"
                )
            )
            exit(1)

        intervals = extract_intervals(task_df, logger=logger)

        intervals_detailed = intervals[["date", "start", "stop", "interval"]].rename(
            columns={
                "date": "Date",
                "start": "Start",
                "stop": "Stop",
                "interval": "Duration",
            }
        )
        print("Log entries:\n")
        print(
            intervals_detailed.to_string(
                index=False,
                formatters={
                    "Start": lambda x: x.strftime("%H:%M:%S"),
                    "Stop": lambda x: x.strftime("%H:%M:%S"),
                    "Duration": lambda x: format_timedelta(
                        timedelta(microseconds=int(x) / 1e3)
                    ),
                },
            )
        )

        print("---")
        print("Daily aggregated:\n")
        intervals_daily = intervals.groupby(by="date")[["interval"]].sum()
        intervals_daily.index.name = "Date"
        intervals_daily = intervals_daily.rename(columns={"interval": "Duration"})
        print(intervals_daily.to_string())

        print(f"---\nTotal: {intervals_detailed['Duration'].sum()}")

    def _read(self) -> None:
        """
        Read data from input file.
        This method uses `pandas.read_csv` to parse the data.
        """
        date_cols = get_datetime_cols_from_schema(self._schema)
        header = [col for col, _ in self._schema]
        try:
            self._log_df = pd.read_csv(
                self._log_fp,
                sep=self._separator,
                parse_dates=date_cols,
                header=None,
                names=header,
            ).sort_values(by=[wc.COL_LOG_DATETIME])
        except pd.errors.EmptyDataError:
            self._log_df = empty_df_from_schema(self._schema)

        self._log_df = pd.concat(
            [self._log_df, extract_date_and_time(self._log_df)], axis=1
        )

    def _persist(self, df: pd.DataFrame, mode="a") -> None:
        cols = [col for col, _ in self._schema]
        df[cols].to_csv(
            self._log_fp, mode=mode, sep=self._separator, index=False, header=False
        )

    def _commit(
        self,
        category: str,
        type_: str,
        log_dt: datetime,
        identifier: str = None,
        force: bool = False,
    ) -> None:
        if type_ not in [wc.TOKEN_START, wc.TOKEN_STOP]:
            raise ValueError(
                f'Type must be one of {", ".join([wc.TOKEN_START, wc.TOKEN_STOP])}'
            )

        commit_dt = datetime.now(timezone.utc).astimezone().replace(microsecond=0)

        # Test if there are running tasks
        if category == wc.TOKEN_SESSION:
            date_mask = self._log_df["date"] == log_dt.date()
            task_mask = self._log_df[wc.COL_CATEGORY] == wc.TOKEN_TASK
            mask = date_mask & task_mask
            active_tasks = get_active_task_ids(self._log_df[mask])
            if len(active_tasks) > 0:
                if not force:
                    msg = self._err_msg_commit_active_tasks.format(
                        active_tasks=active_tasks
                    )
                    sys.stderr.write(msg)
                    sys.exit(1)
                else:
                    for task_id in active_tasks:
                        self._commit(wc.TOKEN_TASK, wc.TOKEN_STOP, log_dt, task_id)

        cols = [col for col, _ in self._schema]
        values = [
            pd.to_datetime(commit_dt),
            pd.to_datetime(log_dt),
            category,
            type_,
            identifier,
        ]

        record = pd.DataFrame(dict(zip(cols, values)), index=[0],)
        record_t = pd.concat([record, extract_date_and_time(record)], axis=1)

        # append record to in-memory log
        self._log_df = pd.concat((self._log_df, record_t))
        # and persist to disk
        self._persist(record_t, mode="a")

        # Because we allow for time offsets sorting is not guaranteed at this point.
        # Update sorting of values in-memory.
        self._log_df = self._log_df.sort_values(by=[wc.COL_LOG_DATETIME])

    def _check_nonempty_or_exit(self, fmt: Optional[str]):
        """
        Tests if the log file has at least a single value.
        Exits with code 0 if no entry is available.
        """
        if self._log_df.shape[0] == 0:
            if fmt is None:
                sys.stderr.write(self._err_msg_empty_log)
            else:
                sys.stdout.write(self._err_msg_empty_log_short)
            sys.exit(0)

    def _filter_date_category_limit_cols(
        self,
        query_date: date,
        filter_category: str = wc.TOKEN_SESSION,
        columns: List[str] = [wc.COL_LOG_DATETIME, wc.COL_TYPE],
    ):
        """
        Filters the worklog DataFrame by query date and category.
        The returned DataFrame only includes the columns listed in the
        `columns` parameter.
        """
        # Extract the day of interest by selecting a subset of the log
        # dataframe that matches the queried day.
        mask = (self._log_df.date == query_date) & (
            self._log_df.category == filter_category
        )
        df = self._log_df[mask]
        df = df[columns]
        return df

    def _add_sentinel(self, query_date: date, df: pd.DataFrame):
        is_active = is_active_session(df)
        ret = df
        if is_active:
            sdt = sentinel_datetime(query_date)
            # attach another row with the current time
            sentinel_df = pd.DataFrame(
                {
                    wc.COL_LOG_DATETIME: pd.to_datetime(sdt.isoformat()),
                    wc.COL_TYPE: wc.TOKEN_STOP,
                },
                index=[0],
            )
            ret = pd.concat((ret, sentinel_df))
            logger.warning(f"Set sentinel stop value: {sdt}")
        return ret

    def _calc_facts(self, df: pd.DataFrame, hours_target: float, hours_max: float):
        shifted_dt = df[wc.COL_LOG_DATETIME].shift(1)
        stop_mask = df[wc.COL_TYPE] == wc.TOKEN_STOP
        total_time = (df[stop_mask][wc.COL_LOG_DATETIME] - shifted_dt[stop_mask]).sum()
        total_time_str = format_timedelta(total_time)

        break_duration = self.auto_break.get_duration(total_time)

        hours_target_dt = timedelta(hours=hours_target) + break_duration
        hours_max_dt = timedelta(hours=hours_max) + break_duration

        now = datetime.now(timezone.utc).astimezone().replace(microsecond=0)
        end_time = now + (hours_target_dt - total_time)
        end_time_str = end_time.strftime("%H:%M:%S")
        remaining_time = max(end_time - now, timedelta(minutes=0))
        remaining_time_str = format_timedelta(remaining_time)
        overtime = max(total_time - hours_target_dt, timedelta(minutes=0))
        overtime_str = format_timedelta(overtime)

        percentage = round(
            total_time.total_seconds() / hours_target_dt.total_seconds() * 100
        )
        percentage_remaining = max(0, 100 - percentage)
        percentage_overtime = max(
            round(
                overtime.total_seconds()
                / (hours_max_dt - hours_target_dt).total_seconds()
                * 100
            ),
            0,
        )
        return dict(
            percentage=percentage,
            end_of_work=end_time_str,
            total_time=total_time_str,
            remaining_time=remaining_time_str,
            remaining_time_short=remaining_time_str[: len("00:00")],
            percentage_remaining=percentage_remaining,
            overtime=overtime_str,
            overtime_short=overtime_str[: len("00:00")],
            percentage_overtime=percentage_overtime,
            break_duration=format_timedelta(break_duration),
        )

    def _aggregate_base(self, mask, keep_cols: List[str] = []):
        df = self._log_df[mask]
        df = df.sort_values([wc.COL_LOG_DATETIME, wc.COL_TYPE])
        shifted_dt = df[wc.COL_LOG_DATETIME].shift(1)
        stop_mask = df[wc.COL_TYPE] == wc.TOKEN_STOP
        agg_time = df[stop_mask][wc.COL_LOG_DATETIME] - shifted_dt[stop_mask]
        ret = df[stop_mask][[wc.COL_LOG_DATETIME] + keep_cols]
        ret["agg_time"] = agg_time
        return ret

    def _aggregate_time(self, mask, resample="D"):
        df = self._aggregate_base(mask, keep_cols=["date"])
        df_day = (
            df.set_index(wc.COL_LOG_DATETIME)
            .resample(resample)
            .sum()
            .reset_index()
            .dropna()
        )
        return df_day

    def _aggregate_tasks(self, mask):
        df = calc_task_durations(
            self._log_df[mask],
            keep_cols=[wc.COL_LOG_DATETIME, wc.COL_TASK_IDENTIFIER, "time"],
        )
        df.rename(columns={"time": "agg_time"}, inplace=True)
        return (
            df.set_index(wc.COL_LOG_DATETIME)
            .groupby(wc.COL_TASK_IDENTIFIER)
            .sum()
            .reset_index()
        )

    def _print_aggregation(self, agg_label, df, cols, col_titles, formatters=None):
        headline = f"Aggregated by {agg_label}:"
        print(headline)
        print("-" * len(headline))
        print(
            df.to_string(
                index=False,
                columns=cols,
                header=col_titles,
                formatters=formatters,
                col_space=20,
            )
        )

        print()



from configparser import ConfigParser
from argparse import ArgumentParser, Namespace
from datetime import date, timedelta
import json

import worklog.constants as wc
from worklog.log import Log
from worklog.utils.time import calc_log_time


def dispatch(
    log: Log, parser: ArgumentParser, cli_args: Namespace, cfg: ConfigParser
) -> None:
    """
    Dispatch request to Log instance based on CLI arguments and
    configuration values.
    """
    if cli_args.subcmd == wc.SUBCMD_COMMIT:
        if cli_args.type in [wc.TOKEN_START, wc.TOKEN_STOP]:
            log.commit(
                wc.TOKEN_SESSION,
                cli_args.type,
                cli_args.offset_minutes,
                cli_args.time,
                force=cli_args.force,
            )
    elif cli_args.subcmd == wc.SUBCMD_TASK:
        if cli_args.type in [wc.TOKEN_START, wc.TOKEN_STOP]:
            if cli_args.type == wc.TOKEN_START and cli_args.auto_close:
                commit_dt = calc_log_time(cli_args.offset_minutes, cli_args.time)
                log.stop_active_tasks(commit_dt)
            log.commit(
                wc.TOKEN_TASK,
                cli_args.type,
                cli_args.offset_minutes,
                cli_args.time,
                identifier=cli_args.id,
            )
        elif cli_args.type == "list":
            log.list_tasks()
        elif cli_args.type == "report":
            log.task_report(cli_args.id)
    elif cli_args.subcmd == wc.SUBCMD_STATUS:
        hours_target = float(cfg.get("workday", "hours_target"))
        hours_max = float(cfg.get("workday", "hours_max"))
        fmt = cli_args.fmt
        query_date = date.today()
        if cli_args.yesterday:
            query_date -= timedelta(days=1)
        elif cli_args.date:
            query_date = cli_args.date.date()
        log.status(hours_target, hours_max, query_date=query_date, fmt=fmt)
    elif cli_args.subcmd == wc.SUBCMD_DOCTOR:
        log.doctor()
    elif cli_args.subcmd == wc.SUBCMD_LOG:
        n = cli_args.number
        no_pager_max_entries = int(cfg.get("worklog", "no_pager_max_entries"))
        use_pager = not cli_args.no_pager and (cli_args.all or n > no_pager_max_entries)
        categories = cli_args.category
        if not cli_args.all:
            log.log(cli_args.number, use_pager, categories)
        else:
            log.log(-1, use_pager, categories)
    elif cli_args.subcmd == wc.SUBCMD_REPORT:
        log.report(cli_args.date_from, cli_args.date_to)



from pandas import DataFrame
import pandas as pd

import worklog.constants as wc

SCHEMA = [
    (wc.COL_COMMIT_DATETIME, "datetime64[ns]",),
    (wc.COL_LOG_DATETIME, "datetime64[ns]",),
    (wc.COL_CATEGORY, "object",),
    (wc.COL_TYPE, "object",),
    (wc.COL_TASK_IDENTIFIER, "object",),
]

SCHEMA_COL_NAMES = [col for col, _ in SCHEMA]
SCHEMA_DATE_COLS = [wc.COL_COMMIT_DATETIME, wc.COL_LOG_DATETIME]


def read_log_sample(id: str) -> DataFrame:
    return pd.read_csv(
        f"worklog/tests/snapshots/{id}.csv",
        sep="|",
        header=None,
        names=SCHEMA_COL_NAMES,
        parse_dates=SCHEMA_DATE_COLS,
    )



import unittest
from unittest.mock import Mock
import pandas as pd
from pandas import DataFrame
from datetime import datetime, date, timedelta, timezone
import logging

import worklog.constants as wc
from worklog.tests.utils import read_log_sample, SCHEMA
from worklog.utils.schema import empty_df_from_schema
from worklog.utils.tasks import (
    _calc_single_task_duration,
    calc_task_durations,
    extract_intervals,
    get_active_task_ids,
    get_all_task_ids_with_duration,
)


class TestTaskIds(unittest.TestCase):
    def test_get_all_task_ids_with_duration_empty(self):
        df = empty_df_from_schema(SCHEMA)
        expected = {}
        actual = get_all_task_ids_with_duration(df)
        self.assertEqual(actual, expected)

    def test_get_all_task_ids_with_duration(self):
        df = read_log_sample("tasks_multiple_nested")
        expected = {
            "task1": timedelta(hours=2),
            "task2": timedelta(minutes=1),
            "task3": timedelta(hours=1, minutes=30),
        }
        actual = get_all_task_ids_with_duration(df)
        self.assertEqual(actual, expected)

    def test_get_active_task_ids_empty(self):
        df = empty_df_from_schema(SCHEMA)
        expected = []
        actual = get_active_task_ids(df)
        self.assertEqual(actual, expected)

    def test_get_active_task_ids_single(self):
        df = read_log_sample("tasks_simple_active")
        expected = ["task1"]
        actual = get_active_task_ids(df)
        self.assertEqual(actual, expected)

    def test_get_active_task_ids_multiple(self):
        df = read_log_sample("tasks_multiple_started")
        expected = ["task1", "task3"]
        actual = get_active_task_ids(df)
        self.assertEqual(actual, expected)


class TestTaskIntervals(unittest.TestCase):
    def test_extract_intervals_empty_dataframe(self):
        df = DataFrame()

        actual = extract_intervals(df)
        expected = DataFrame(columns=["date", "start", "stop", "interval"])

        pd.testing.assert_frame_equal(actual, expected)

    def test_extract_intervals_valid(self):
        mock_logger = Mock(logging.Logger)
        df = read_log_sample("tasks_simple_ordered")

        actual = extract_intervals(df, logger=mock_logger)
        expected = DataFrame(
            {
                "date": [date(2020, 1, 1)],
                "start": [datetime(2020, 1, 1, 0, tzinfo=timezone.utc)],
                "stop": [datetime(2020, 1, 1, 1, tzinfo=timezone.utc)],
                "interval": [timedelta(hours=1)],
            }
        )

        pd.testing.assert_frame_equal(actual, expected)
        mock_logger.error.assert_not_called()

    def test_extract_intervals_invalid_start_missing(self):
        mock_logger = Mock(logging.Logger)
        df = read_log_sample("tasks_start_missing")

        actual = extract_intervals(df, logger=mock_logger)
        expected = DataFrame(columns=["date", "start", "stop", "interval"])

        pd.testing.assert_frame_equal(actual, expected)
        mock_logger.error.assert_called_once_with("No start entry found. Skip entry.")

    def test_extract_intervals_invalid_stop_missing(self):
        mock_logger = Mock(logging.Logger)
        df = read_log_sample("tasks_open_interval")

        actual = extract_intervals(df, logger=mock_logger)
        expected = DataFrame(
            {
                "date": [date(2020, 1, 1)],
                "start": [datetime(2020, 1, 1, 0, tzinfo=timezone.utc)],
                "stop": [datetime(2020, 1, 1, 1, tzinfo=timezone.utc)],
                "interval": [timedelta(hours=1)],
            }
        )

        pd.testing.assert_frame_equal(actual, expected)
        mock_logger.error.assert_called_once_with(
            "Start entry at 2020-01-01 01:30:00+00:00 has no stop entry. Skip entry."
        )

    def test_extract_intervals_invalid_unknown_type(self):
        mock_logger = Mock(logging.Logger)
        df = read_log_sample("tasks_invalid_type")

        actual = extract_intervals(df, logger=mock_logger)
        expected = DataFrame(
            {
                "date": [date(2020, 1, 1)],
                "start": [datetime(2020, 1, 1, 0, tzinfo=timezone.utc)],
                "stop": [datetime(2020, 1, 1, 1, tzinfo=timezone.utc)],
                "interval": [timedelta(hours=1)],
            }
        )

        pd.testing.assert_frame_equal(actual, expected)
        mock_logger.error.assert_called_once_with(
            "Found unknown type 'unknown'. Skip entry."
        )


class TestTaskDuration(unittest.TestCase):
    def test_calc_task_durations_ordered(self):
        df = read_log_sample("tasks_simple_ordered")
        expected = DataFrame(
            {wc.COL_TASK_IDENTIFIER: ["foo"], "time": [timedelta(hours=1)]}, index=[1]
        )
        actual = _calc_single_task_duration(
            df, keep_cols=[wc.COL_TASK_IDENTIFIER, "time"]
        )
        pd.testing.assert_frame_equal(actual, expected)

    def test_calc_task_durations_open_interval(self):
        df = read_log_sample("tasks_open_interval")
        expected = DataFrame(
            {wc.COL_TASK_IDENTIFIER: ["foo"], "time": [timedelta(hours=1)]}, index=[1]
        )
        actual = _calc_single_task_duration(
            df, keep_cols=[wc.COL_TASK_IDENTIFIER, "time"]
        )
        pd.testing.assert_frame_equal(actual, expected)

    def calc_task_durations_multiple_ordered(self):
        df = read_log_sample("tasks_multiple_ordered")
        expected = DataFrame(
            {
                wc.COL_TASK_IDENTIFIER: ["bar", "foo"],
                "time": [timedelta(minutes=30), timedelta(hours=1)],
            },
            index=[3, 1],
        )
        actual = calc_task_durations(df)
        pd.testing.assert_frame_equal(actual, expected)

    def test_calc_task_durations_multiple_nested(self):
        """
        Test if the implementation works if tasks are nested.
        """
        df = read_log_sample("tasks_multiple_nested")
        expected = DataFrame(
            {
                wc.COL_TASK_IDENTIFIER: ["task1", "task2", "task3"],
                "time": [
                    timedelta(hours=2),
                    timedelta(minutes=1),
                    timedelta(hours=1, minutes=30),
                ],
            },
            index=pd.MultiIndex.from_tuples(
                [("task1", 4), ("task2", 3), ("task3", 5)],
                names=[wc.COL_TASK_IDENTIFIER, None],
            ),
        )
        actual = calc_task_durations(df)
        pd.testing.assert_frame_equal(actual, expected)

    def test_calc_task_durations_multiple_unordered(self):
        """
        Test if implementation works if the log entries are in the wrong
        order. In this case a stop entry is listed before a start entry
        although the order of the logged time is fine.
        """
        df = read_log_sample("tasks_multiple_nested_unordered")
        expected = DataFrame(
            {
                wc.COL_TASK_IDENTIFIER: ["task1", "task2"],
                "time": [timedelta(hours=2), timedelta(minutes=1)],
            },
            index=pd.MultiIndex.from_tuples(
                [("task1", 1), ("task2", 0)], names=[wc.COL_TASK_IDENTIFIER, None],
            ),
        )
        actual = calc_task_durations(df)
        pd.testing.assert_frame_equal(actual, expected)



from typing import Optional
from datetime import datetime, timedelta, timezone
import pandas as pd

import worklog.constants as wc


def _get_or_update_dt(dt: datetime, time: str):
    try:
        h_time = datetime.strptime(time, "%H:%M")
        hour, minute = h_time.hour, h_time.minute
        return dt.replace(hour=hour, minute=minute, second=0)
    except ValueError:
        h_time = datetime.fromisoformat(time).replace(second=0)
        if h_time.tzinfo is None:
            # Set local timezone if not defined explicitly.
            h_time = h_time.replace(tzinfo=wc.LOCAL_TIMEZONE)
        return h_time


def calc_log_time(offset_min: int = 0, time: Optional[str] = None) -> datetime:
    """
    Calculates the log time based on the current timestamp and either an
    offset or a time correction.
    """
    my_date = datetime.now(timezone.utc).astimezone().replace(microsecond=0)
    my_date = my_date + timedelta(minutes=offset_min)

    if time is not None:
        my_date = _get_or_update_dt(my_date, time)

    return my_date


def extract_date_and_time(
    df: pd.DataFrame, source_col: str = wc.COL_LOG_DATETIME
) -> pd.DataFrame:
    """
    Extracts date and time information from a given pandas DataFrame.
    By default the source column is `log_dt`.
    """
    date: pd.Series = df[source_col].apply(lambda x: x.date)
    time: pd.Series = df[source_col].apply(lambda x: x.time)
    return pd.DataFrame(dict(date=date, time=time),)



import os
from configparser import ConfigParser
from io import StringIO
import json
import pkg_resources

from worklog.breaks import AutoBreak
import worklog.constants as wc
from worklog.log import Log
from worklog.parser import get_arg_parser
from worklog.utils.time import calc_log_time
from worklog.utils.logger import configure_logger
from worklog.dispatcher import dispatch


try:
    __version__ = pkg_resources.get_distribution("dcs-" + __name__).version
except Exception:
    __version__ = "unknown"


def run() -> None:
    """ Main method """
    logger = configure_logger()
    parser = get_arg_parser()

    cli_args = parser.parse_args()
    log_level = wc.LOG_LEVELS[min(cli_args.verbosity, len(wc.LOG_LEVELS) - 1)]
    logger.setLevel(log_level)

    logger.debug(f"Parsed CLI arguments: {cli_args}")
    logger.debug(f"Path to config files: {wc.CONFIG_FILES}")

    if cli_args.subcmd is None:
        parser.print_help()
        return

    cfg = ConfigParser()
    cfg.read(wc.CONFIG_FILES)

    with StringIO() as ss:
        cfg.write(ss)
        ss.seek(0)
        logger.debug(f"Config content:\n{ss.read()}\nEOF")

    worklog_fp = os.path.expanduser(cfg.get("worklog", "path"))
    log = Log(worklog_fp)

    limits = json.loads(cfg.get("workday", "auto_break_limit_minutes"))
    durations = json.loads(cfg.get("workday", "auto_break_duration_minutes"))
    log.auto_break = AutoBreak(limits, durations)

    dispatch(log, parser, cli_args, cfg)


if __name__ == "__main__":
    run()



import unittest
from unittest.mock import patch
from io import StringIO
from argparse import ArgumentParser, ArgumentError, ArgumentTypeError
from datetime import datetime

from worklog.constants import LOCAL_TIMEZONE
from worklog.parser import (
    get_arg_parser,
    _positive_int,
    _combined_month_or_day_or_week_parser,
    _year_month_day_parser,
    _year_month_parser,
    _calendar_week_parser,
)


class TestParser(unittest.TestCase):
    def setUp(self):
        self.parser = get_arg_parser()

    def tearDown(self):
        self.parser = None

    def test_positive_int_with_pos_int(self):
        expected = 5
        actual = _positive_int("5")
        self.assertEqual(expected, actual)

    def test_positive_int_with_neg_int(self):
        with self.assertRaises(ArgumentTypeError):
            _positive_int("-5")

    def test_combined_month_or_day_or_week_parser_value_unknown(self):
        with self.assertRaises(ArgumentTypeError):
            _combined_month_or_day_or_week_parser("foobar")

    @patch("worklog.parser._year_month_parser")
    def test_combined_month_or_day_or_week_parser_ym(self, ym_mock):
        _combined_month_or_day_or_week_parser("2020-01")
        ym_mock.assert_called_with("2020-01")

    @patch("worklog.parser._year_month_day_parser")
    def test_combined_month_or_day_or_week_parser_ymd(self, ymd_mock):
        _combined_month_or_day_or_week_parser("2020-01-01")
        ymd_mock.assert_called_with("2020-01-01")

    @patch("worklog.parser._calendar_week_parser")
    def test_combined_month_or_day_or_week_parser_cw(self, cw_mock):
        _combined_month_or_day_or_week_parser("2020-W01")
        cw_mock.assert_called_with("2020-W01")

    def test_ym_parser_invalid_value(self):
        with self.assertRaises(ArgumentTypeError):
            _year_month_parser("foobar")

    def test_ym_parser_ym(self):
        expected = datetime(year=2020, month=1, day=1, tzinfo=LOCAL_TIMEZONE)
        actual = _year_month_parser("2020-01")
        self.assertEqual(expected, actual)

    def test_ymd_parser_invalid_value(self):
        with self.assertRaises(ArgumentTypeError):
            _year_month_day_parser("foobar")

    def test_ymd_parser_ymd(self):
        expected = datetime(year=2020, month=1, day=2, tzinfo=LOCAL_TIMEZONE)
        actual = _year_month_day_parser("2020-01-02")
        self.assertEqual(expected, actual)

    def test_cw_parser_invalid_value(self):
        with self.assertRaises(ArgumentTypeError):
            _calendar_week_parser("foobar")

    def test_cw_parser_ymd(self):
        expected = datetime(year=2019, month=12, day=30, tzinfo=LOCAL_TIMEZONE)
        actual = _calendar_week_parser("2020-W00")
        self.assertEqual(expected, actual)
        expected = datetime(year=2020, month=1, day=6, tzinfo=LOCAL_TIMEZONE)
        actual = _calendar_week_parser("2020-W01")
        self.assertEqual(expected, actual)

    @patch("sys.stderr", new_callable=StringIO)
    def test_missing_subcmd_throws_and_exits(self, mock_err):
        with self.assertRaises(SystemExit):
            with self.assertRaises(Exception):
                self.parser.parse_args()

        self.assertIn("invalid choice", mock_err.getvalue())

    def test_subcmd_status(self):
        argv = ["status"]
        cli_args = self.parser.parse_args(argv)

        self.assertEqual(cli_args.subcmd, "status")
        self.assertFalse(cli_args.yesterday)
        self.assertIsNone(cli_args.fmt)

    def test_subcmd_status_yesterday(self):
        argv = ["status", "--yesterday"]
        cli_args = self.parser.parse_args(argv)

        self.assertEqual(cli_args.subcmd, "status")
        self.assertTrue(cli_args.yesterday)

    def test_subcmd_status_fmt(self):
        argv = ["status", "--fmt", "{position}"]
        cli_args = self.parser.parse_args(argv)

        self.assertEqual(cli_args.subcmd, "status")
        self.assertEqual(cli_args.fmt, "{position}")

    def test_subcmd_doctor(self):
        argv = ["doctor"]
        cli_args = self.parser.parse_args(argv)

        self.assertEqual(cli_args.subcmd, "doctor")

    def test_subcmd_log(self):
        argv = ["log"]
        cli_args = self.parser.parse_args(argv)

        self.assertEqual(cli_args.subcmd, "log")
        self.assertEqual(cli_args.number, 10)
        self.assertFalse(cli_args.all)

    def test_subcmd_log_with_pos_number(self):
        argv = ["log", "-n", "5"]
        cli_args = self.parser.parse_args(argv)

        self.assertEqual(cli_args.subcmd, "log")
        self.assertEqual(cli_args.number, 5)

    @patch("sys.stderr", new_callable=StringIO)
    def test_subcmd_log_with_neg_number(self, mock_err):
        argv = ["log", "-n", "-5"]

        with self.assertRaises(SystemExit):
            with self.assertRaises(Exception):
                self.parser.parse_args(argv)

        self.assertIn(
            "argument -n/--number: -5 is not a positive int value", mock_err.getvalue()
        )

    def test_subcmd_log_all(self):
        argv = ["log", "--all"]
        cli_args = self.parser.parse_args(argv)

        self.assertEqual(cli_args.subcmd, "log")
        self.assertTrue(cli_args.all)




import unittest
from unittest.mock import patch
import logging
from datetime import datetime, date, timezone

from worklog.utils.schema import empty_df_from_schema
from worklog.utils.session import (
    check_order_session,
    sentinel_datetime,
    is_active_session,
)
from worklog.tests.utils import read_log_sample


class TestSessionOrder(unittest.TestCase):
    def test_check_order_session(self):
        schema = [
            ("log_dt", "datetime64[ns]",),
            ("category", "object",),
            ("type", "object",),
        ]

        rows = [
            {
                "log_dt": datetime(2020, 1, 1, 0, 0, 0, 0, timezone.utc),
                "category": "session",
                "type": "start",
            },
            {
                "log_dt": datetime(2020, 1, 1, 1, 0, 0, 0, timezone.utc),
                "category": "session",
                "type": "stop",
            },
            {
                "log_dt": datetime(2020, 1, 1, 2, 0, 0, 0, timezone.utc),
                "category": "session",
                "type": "stop",
            },
        ]

        logger = logging.getLogger("test_check_order_session")
        df = empty_df_from_schema(schema)

        # Positive case
        df1 = df.append(rows[:2], ignore_index=True)
        df1["date"] = df1["log_dt"].apply(lambda x: x.date())
        df1["time"] = df1["log_dt"].apply(lambda x: x.time())
        with patch.object(logger, "error") as mock_error:
            check_order_session(df1, logger)
            mock_error.assert_not_called()

        # Two stop entries after each other -> Error!
        df2 = df.append(rows[:3], ignore_index=True)
        df2["date"] = df2["log_dt"].apply(lambda x: x.date())
        df2["time"] = df2["log_dt"].apply(lambda x: x.time())
        with patch.object(logger, "error") as mock_error:
            check_order_session(df2, logger)
            mock_error.assert_called_with(
                '"session" entries on date 2020-01-01 are not ordered correctly.'
            )

        # First entry is a 'stop' entry -> Error!
        df3 = df.append(rows[1:2], ignore_index=True)
        df3["date"] = df3["log_dt"].apply(lambda x: x.date())
        df3["time"] = df3["log_dt"].apply(lambda x: x.time())
        with patch.object(logger, "error") as mock_error:
            check_order_session(df3, logger)
            mock_error.assert_called_with(
                'First entry of type "session" on date 2020-01-01 is not "start".'
            )

        # Last entry is 'start' and 'stop' entry is missing -> Error!
        df4 = df.append(rows[0:1], ignore_index=True)
        df4["date"] = df4["log_dt"].apply(lambda x: x.date())
        df4["time"] = df4["log_dt"].apply(lambda x: x.time())
        with patch.object(logger, "error") as mock_error:
            check_order_session(df4, logger)
            mock_error.assert_called_with("Date 2020-01-01 has no stop entry.")


class TestSentinelEntries(unittest.TestCase):
    @patch("worklog.utils.session.datetime")
    def test_sentinel_datetime(self, mock_datetime):
        mock_datetime.now.return_value = datetime(2020, 1, 2, 1, 33, 7, 0, timezone.utc)
        mock_datetime.side_effect = lambda *args, **kwargs: datetime(*args, **kwargs)

        # Date is in the past -> Sentinal value is last second on this date
        target_date1 = date(2019, 1, 1)
        actual_1 = sentinel_datetime(target_date1, tzinfo=timezone.utc)
        self.assertEqual(actual_1.isoformat(), "2019-01-01T23:59:59+00:00")

        # Date is on the same day as today -> Sentinel value is datetime.now()
        target_date2 = date(2020, 1, 2)
        actual_2 = sentinel_datetime(target_date2, tzinfo=timezone.utc)
        self.assertEqual(actual_2.isoformat(), "2020-01-02T01:33:07+00:00")

        # Date is yesterday -> Sentinel value is the last second on this date
        target_date3 = date(2020, 1, 1)
        actual_3 = sentinel_datetime(target_date3, tzinfo=timezone.utc)
        self.assertEqual(actual_3.isoformat(), "2020-01-01T23:59:59+00:00")

        # Date is in the future -> Raise error
        with self.assertRaises(ValueError):
            target_date4 = date(2020, 1, 3)
            sentinel_datetime(target_date4, tzinfo=timezone.utc)


class TestSessionActivity(unittest.TestCase):
    def test_inactive_session(self):
        df = read_log_sample("session_simple")
        actual = is_active_session(df)

        self.assertFalse(actual)

    def test_active_session(self):
        df = read_log_sample("session_simple_open")
        actual = is_active_session(df)

        self.assertTrue(actual)



import unittest
from datetime import timedelta

from worklog.breaks import AutoBreak


class TestBreaks(unittest.TestCase):
    def test_zero(self):
        td = timedelta(minutes=0)
        auto_break = AutoBreak()

        expected = 0
        actual = auto_break.get_duration(td).total_seconds() // 60

        self.assertEqual(expected, actual)

    def test_nonequal_list_lens(self):
        limits = [0, 360]  # len: 2
        durations = [15]  # len: 1

        with self.assertRaises(ValueError):
            auto_break = AutoBreak(limits, durations)

    def test_in_first_interval(self):
        limits = [0, 360]  # in minutes
        durations = [15, 45]  # in minutes
        auto_break = AutoBreak(limits, durations)

        td = timedelta(minutes=5)
        expected = 15

        actual = auto_break.get_duration(td).total_seconds() // 60

        self.assertEqual(expected, actual)

    def test_above_first_interval(self):
        limits = [0, 360]  # in minutes
        durations = [15, 45]  # in minutes
        auto_break = AutoBreak(limits, durations)

        td = timedelta(minutes=361)
        expected = 45
        actual = auto_break.get_duration(td).total_seconds() // 60

        self.assertEqual(expected, actual)

    def test_in_mid_interval(self):
        limits = [0, 200, 360]  # in minutes
        durations = [15, 20, 45]  # in minutes
        auto_break = AutoBreak(limits, durations)

        td = timedelta(minutes=201)
        expected = 20
        actual = auto_break.get_duration(td).total_seconds() // 60

        self.assertEqual(expected, actual)



from typing import List, Optional
from pandas import DataFrame
from datetime import datetime
import logging

import worklog.constants as wc


def calc_task_durations(
    df: DataFrame, keep_cols: List[str] = [wc.COL_TASK_IDENTIFIER, "time"]
) -> DataFrame:
    return df.groupby(wc.COL_TASK_IDENTIFIER).apply(
        lambda group: _calc_single_task_duration(group, keep_cols=keep_cols)
    )


def _calc_single_task_duration(df: DataFrame, keep_cols: List[str] = []):
    """
    Calculate the duration of a single task. The given DataFrame can consist
    of many log entries but all must have the same task identifier.
    """
    # TODO: Sanity check missing: Each start row should have a
    # corresponding stop row
    # TODO: Generalize this idea to other parts where time is calculated
    df = df.sort_values([wc.COL_LOG_DATETIME, wc.COL_TYPE])
    stop_mask = df[wc.COL_TYPE] == wc.TOKEN_STOP
    shifted_dt = df[wc.COL_LOG_DATETIME].shift(1)
    s_time = df[stop_mask][wc.COL_LOG_DATETIME] - shifted_dt[stop_mask]

    df_result = df[stop_mask].copy()
    df_result["time"] = s_time
    df_result = df_result.sort_values(by=[wc.COL_TASK_IDENTIFIER])
    return df_result[keep_cols]


def get_all_task_ids_with_duration(df: DataFrame):
    df = df[[wc.COL_LOG_DATETIME, wc.COL_TYPE, wc.COL_TASK_IDENTIFIER]].sort_values(
        by=[wc.COL_LOG_DATETIME]
    )

    df_h = calc_task_durations(df)
    if df_h.empty:
        return {}
    s = df_h["time"]
    s.index = df_h.index.map(lambda k: k[0])
    return s.to_dict()


def get_active_task_ids(df: DataFrame):
    """
    Returns a list of active tasks.
    Note: This method can only handle DataFrames that have entries where the
    category is set to 'task'.
    """
    df = df[[wc.COL_LOG_DATETIME, wc.COL_TYPE, wc.COL_TASK_IDENTIFIER]].sort_values(
        by=[wc.COL_LOG_DATETIME]
    )

    df_grouped = df.groupby(wc.COL_TASK_IDENTIFIER).tail(1)
    return sorted(
        df_grouped[df_grouped[wc.COL_TYPE] == wc.TOKEN_START][
            wc.COL_TASK_IDENTIFIER
        ].unique()
    )


def extract_intervals(
    df: DataFrame, logger: Optional[logging.Logger] = None,
):
    def log_error(msg):
        if logger:
            logger.error(msg)

    intervals = []
    last_start: Optional[datetime] = None
    for i, row in df.iterrows():
        if row[wc.COL_TYPE] == wc.TOKEN_START:
            if last_start is not None:
                log_error(f"Start entry at {last_start} has no stop entry. Skip entry.")
            last_start = row[wc.COL_LOG_DATETIME]
        elif row[wc.COL_TYPE] == wc.TOKEN_STOP:
            if last_start is None:
                log_error("No start entry found. Skip entry.")
                continue  # skip this entry
            td = row[wc.COL_LOG_DATETIME] - last_start
            d = last_start.date()
            intervals.append(
                {
                    "date": d,
                    "start": last_start,
                    "stop": row[wc.COL_LOG_DATETIME],
                    "interval": td,
                }
            )
            last_start = None
        else:
            log_error(f"Found unknown type '{row['type']}'. Skip entry.")
            continue
    if last_start is not None:
        log_error(f"Start entry at {last_start} has no stop entry. Skip entry.")

    return DataFrame(intervals, columns=["date", "start", "stop", "interval"])



import unittest
from datetime import timedelta
import numpy as np
import pandas as pd

from worklog.utils.formatting import format_timedelta


class TestTimeFormatting(unittest.TestCase):
    def test_invalid_value_none(self):
        expected = "00:00:00"

        td = None
        actual = format_timedelta(td)
        self.assertEqual(expected, actual)

    def test_invalid_value_nat(self):
        expected = "00:00:00"
        td = pd.NaT
        actual = format_timedelta(td)
        self.assertEqual(expected, actual)

    def test_invalid_value_nan(self):
        expected = "00:00:00"
        td = np.nan
        actual = format_timedelta(td)
        self.assertEqual(expected, actual)

    def test_default_behaviour(self):
        td = timedelta(hours=1, minutes=5, seconds=30)
        expected = "01:05:30"
        actual = format_timedelta(td)
        self.assertEqual(expected, actual)

    def test_format_timedelta_nat(self):
        """It should not break if a numpy not-a-time value is passed."""
        actual = format_timedelta(np.timedelta64("nAt"))
        expected = "00:00:00"

        self.assertEqual(actual, expected)

    def test_format_timedelta_valid(self):
        actual = format_timedelta(np.timedelta64(100, "s"))
        expected = "00:01:40"

        self.assertEqual(actual, expected)

    def test_format_timedelta_valid_days(self):
        actual = format_timedelta(np.timedelta64(2, "D"))
        expected = "48:00:00"

        self.assertEqual(actual, expected)

        actual = format_timedelta(np.timedelta64(20, "D"))
        expected = "480:00:00"

        self.assertEqual(actual, expected)



from typing import List
from datetime import timedelta


class AutoBreak(object):
    _durations: List[int] = []
    _limits: List[int] = []

    def __init__(self, limits: List[int] = [], durations: List[int] = []):
        if len(limits) != len(durations):
            raise ValueError("limits and durations must have the same shape.")

        self._limits = limits
        self._durations = durations

    @property
    def active(self):
        return len(self._durations) > 0

    def get_duration(self, td: timedelta):
        """Calculate the break duration for a given timedelta."""
        if not self.active:
            return timedelta(minutes=0)

        full_minutes: int = int(td.total_seconds()) // 60

        break_duration = 0

        for i, limit in enumerate(self._limits):
            if limit > full_minutes:
                break
            break_duration = self._durations[i]

        return timedelta(minutes=break_duration)




import unittest
from unittest.mock import patch, Mock
from datetime import date, time, datetime, timezone
from pandas import DataFrame
import pandas as pd

import worklog.constants as wc
from worklog.utils.time import _get_or_update_dt, calc_log_time, extract_date_and_time
from worklog.tests.utils import read_log_sample


class TestDatetimeManipulation(unittest.TestCase):
    def test_get_or_update_dt_with_time(self):
        dt = datetime(2020, 1, 1, 0, 0, 0, tzinfo=wc.LOCAL_TIMEZONE)

        actual = _get_or_update_dt(dt, "05:06")
        expected = datetime(2020, 1, 1, 5, 6, tzinfo=wc.LOCAL_TIMEZONE)

        actual = _get_or_update_dt(dt, "14:06")
        expected = datetime(2020, 1, 1, 14, 6, tzinfo=wc.LOCAL_TIMEZONE)

        self.assertEqual(actual, expected)

    def test_get_or_update_dt_with_iso_datestr(self):
        dt = datetime(2020, 1, 1, 0, 0, 0, tzinfo=wc.LOCAL_TIMEZONE)

        actual = _get_or_update_dt(dt, "1999-12-13")
        expected = datetime(1999, 12, 13, 0, 0, tzinfo=wc.LOCAL_TIMEZONE)

        actual = _get_or_update_dt(dt, "1999-12-13T01:12")
        expected = datetime(1999, 12, 13, 1, 12, tzinfo=wc.LOCAL_TIMEZONE)

        actual = _get_or_update_dt(dt, "2020-02-03 04:05")
        expected = datetime(2020, 2, 3, 4, 5, tzinfo=wc.LOCAL_TIMEZONE)

        actual = _get_or_update_dt(dt, "2020-02-03 14:05:03")
        expected = datetime(2020, 2, 3, 14, 5, tzinfo=wc.LOCAL_TIMEZONE)

        actual = _get_or_update_dt(dt, "2020-02-03 04:05:03+00:00")
        expected = datetime(2020, 2, 3, 4, 5, tzinfo=timezone.utc)

        self.assertEqual(actual, expected)

    @patch(
        "worklog.utils.time.datetime",
        Mock(now=Mock(return_value=datetime(2020, 1, 1, tzinfo=timezone.utc))),
    )
    def test_calc_log_time_no_corrections(self):

        expected = datetime(2020, 1, 1, tzinfo=timezone.utc)
        actual = calc_log_time()

        self.assertEqual(actual, expected)

    @patch(
        "worklog.utils.time.datetime",
        Mock(now=Mock(return_value=datetime(2020, 1, 1, tzinfo=timezone.utc))),
    )
    @patch("worklog.utils.time._get_or_update_dt")
    def test_calc_log_time_with_time_corrections(self, mock):

        dt = datetime(2020, 1, 1, tzinfo=timezone.utc)
        actual = calc_log_time(time="20:21")

        mock.assert_called_with(dt, "20:21")

    @patch(
        "worklog.utils.time.datetime",
        Mock(now=Mock(return_value=datetime(2020, 1, 1, tzinfo=timezone.utc))),
    )
    @patch("worklog.utils.time._get_or_update_dt")
    def test_calc_log_time_with_offset_minutes_corrections(self, mock):

        expected = datetime(2020, 1, 1, 0, 10, tzinfo=timezone.utc)
        actual = calc_log_time(offset_min=10)

        mock.assert_not_called()
        self.assertEqual(actual, expected)


class TestDateTimeExtraction(unittest.TestCase):
    def test_extraction(self):
        df = read_log_sample("session_simple")

        expected = DataFrame(
            {
                "date": [date(2020, 1, 1), date(2020, 1, 1)],
                "time": [time(0, 0, 0), time(1, 0, 0)],
            }
        )
        actual = extract_date_and_time(df)

        pd.testing.assert_frame_equal(actual, expected)



import unittest
from unittest.mock import patch

from worklog.utils.pager import get_pager


class TestPager(unittest.TestCase):
    @patch("os.getenv")
    @patch("shutil.which")
    def test_get_pager_no_less_and_env_unset(self, mock_which, mock_getenv):
        mock_which.side_effect = ["/path/to/more", None]
        mock_getenv.side_effect = lambda _, default_value: default_value

        actual = get_pager()
        expected = "/path/to/more"

        self.assertEqual(actual, expected)

    @patch("os.getenv")
    @patch("shutil.which")
    def test_get_pager_no_less_and_env_set(self, mock_which, mock_getenv):
        mock_which.side_effect = ["/path/to/more", None]
        mock_getenv.return_value = "/path/to/less"

        actual = get_pager()
        expected = "/path/to/less"

        self.assertEqual(actual, expected)

    @patch("os.getenv")
    @patch("shutil.which")
    def test_get_pager_has_less_and_env_unset(self, mock_which, mock_getenv):
        mock_which.side_effect = ["/path/to/more", "/path/to/less"]
        mock_getenv.side_effect = lambda _, default_value: default_value

        actual = get_pager()
        expected = "/path/to/less"

        self.assertEqual(actual, expected)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    requirements = fh.readlines()

with open("requirements_develop.txt", "r") as fh:
    requirements_develop = fh.readlines()

setuptools.setup(
    name="dcs-worklog",
    version="0.0.31",
    author="Fabian Mueller",
    author_email="repository@dotcs.me",
    description="Simple CLI tool to log work and projects",
    license="GPLv3",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dotcs/worklog",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=requirements,
    extras_require={"develop": requirements_develop},
    entry_points={"console_scripts": ["wl=worklog:run",]},
)



from typing import List, Iterable, Tuple, Dict, Optional, Union
from pandas import DataFrame, Series  # type: ignore
import logging
import sys
import argparse
import os
from functools import reduce
from datetime import datetime, date, timezone, timedelta, tzinfo
import shutil
from math import floor
import numpy as np

import worklog.constants as wc




from typing import List, Optional
import logging
import os
from datetime import datetime, timezone, tzinfo

LOG_FORMAT: str = logging.BASIC_FORMAT
LOG_LEVELS: List[int] = [logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]

CONFIG_FILES: List[str] = [
    os.path.join(os.path.abspath(os.path.dirname(__file__)), "config.cfg"),
    os.path.expanduser("~/.config/worklog/config"),
]

LOCAL_TIMEZONE: Optional[tzinfo] = datetime.now(timezone.utc).astimezone().tzinfo

DEFAULT_LOGGER_NAME = "worklog"

SUBCMD_COMMIT = "commit"
SUBCMD_DOCTOR = "doctor"
SUBCMD_TASK = "task"
SUBCMD_STATUS = "status"
SUBCMD_LOG = "log"
SUBCMD_REPORT = "report"

COL_COMMIT_DATETIME = "commit_dt"
COL_LOG_DATETIME = "log_dt"
COL_CATEGORY = "category"
COL_TYPE = "type"
COL_TASK_IDENTIFIER = "identifier"

TOKEN_START = "start"
TOKEN_STOP = "stop"
TOKEN_SESSION = "session"
TOKEN_TASK = "task"




from typing import Iterable, Tuple, Dict, List
from pandas import DataFrame, Series
from functools import reduce


def empty_df_from_schema(schema: Iterable[Tuple[str, str]]) -> DataFrame:
    def reducer(acc: Dict, x: Tuple[str, str]):
        acc[x[0]] = Series(dtype=x[1])
        return acc

    return DataFrame(reduce(reducer, schema, {}))


def get_datetime_cols_from_schema(schema: Iterable[Tuple[str, str]]) -> List[str]:
    def reducer(acc: List, x: Tuple[str, str]):
        if "datetime" in x[1]:
            acc.append(x[0])
        return acc

    return reduce(reducer, schema, [])



import unittest
import sys
import logging

from worklog.utils.logger import configure_logger
import worklog.constants as wc


class TestLogging(unittest.TestCase):
    def test_configure_logger_name(self):
        logger = configure_logger()
        name = logger.name
        self.assertEqual(name, wc.DEFAULT_LOGGER_NAME)

    def test_configure_logger_handlers(self):
        logger = configure_logger()

        has_handlers = logger.hasHandlers()
        self.assertTrue(has_handlers)

        sys_handler = logger.handlers[0]
        self.assertEqual(sys_handler.stream, sys.stdout)

    def test_configure_logger_default_log_level(self):
        logger = configure_logger()
        log_level = logger.level
        self.assertEqual(log_level, logging.INFO)
